from typing import Any, Dict, List, Literal

from lpsim.server.query import query_one
from lpsim.server.consts import IconType
from lpsim.server.status.character_status.base import (
    PrepareCharacterStatus,
    UsageCharacterStatus,
)
from lpsim.server.status.team_status.base import UsageTeamStatus
from lpsim.utils.class_registry import register_class
from lpsim.server.match import Match
from lpsim.utils.desc_registry import DescDictType
from lpsim.server.modifiable_values import DamageIncreaseValue, DamageValue
from lpsim.server.event import ReceiveDamageEventArguments, SkillEndEventArguments
from lpsim.server.action import Actions, CreateObjectAction, MakeDamageAction
from lpsim.server.struct import Cost
from lpsim.consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from lpsim.server.character.character_base import (
    AOESkillBase,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Character status.


# Round status, will last for several rounds and disappear
class HeirToTheAncientSeasAuthorityStatus_4_5(UsageCharacterStatus):
    name: Literal[
        "Heir to the Ancient Sea's Authority"
    ] = "Heir to the Ancient Sea's Authority"
    version: Literal["4.5"] = "4.5"
    usage: int = 2
    max_usage: int = 2
    icon_type: IconType = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When this character attack, increase one damage.
        """
        if self.usage <= 0:
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            return value
        value.damage += 1
        self.usage -= 1
        return value


# Usage status, will not disappear until usage is 0
class SourcewaterDroplet_4_5(UsageTeamStatus):
    name: Literal["Sourcewater Droplet"] = "Sourcewater Droplet"
    version: Literal["4.5"] = "4.5"
    usage: int = 1
    max_usage: int = 3
    icon_type: IconType = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[Actions]:
        # when Neuvillette uses As Water Seeks Equilibrium, heal 2 hp and prepare skill
        who = query_one(event.action.position, match, "self")
        if (
            who is not None
            and who.name == "Neuvillette"
            and event.action.skill_type == "NORMAL_ATTACK"
            and who.position.player_idx == self.position.player_idx
        ):
            self.usage -= 1
            return [
                MakeDamageAction(
                    damage_value_list=[
                        DamageValue.create_heal(self.position, who.position, -2, Cost())
                    ]
                ),
                CreateObjectAction(
                    object_name="Equitable Judgment",
                    object_position=who.position.set_area(
                        ObjectPositionType.CHARACTER_STATUS
                    ),
                    object_arguments={},
                ),
            ]
        return []


# class IncinerationDrive_4_1(PrepareCharacterStatus):
class EquitableJudgment_4_5(PrepareCharacterStatus):
    name: Literal["Equitable Judgment"] = "Equitable Judgment"
    version: Literal["4.5"] = "4.5"
    character_name: Literal["Neuvillette"] = "Neuvillette"
    skill_name: Literal["Equitable Judgment"] = "Equitable Judgment"


# Skills


class EquitableJudgment(ElementalNormalAttackBase):
    name: Literal["Equitable Judgment"] = "Equitable Judgment"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        char: CharacterBase = self.query_one(match, "self")  # type: ignore
        action = self.attack_opposite_active(match, self.damage, self.damage_type)
        if char.hp >= 6:
            action.damage_value_list[0].damage += 1
            attack_self_action = self.attack_self(match, 1)
            action.damage_value_list.append(attack_self_action.damage_value_list[0])
        return [action]


class OTearsIShallRepay(ElementalSkillBase):
    name: Literal["O Tears, I Shall Repay"] = "O Tears, I Shall Repay"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_team_status("Sourcewater Droplet")
        ]


class OTidesIHaveReturned(ElementalBurstBase, AOESkillBase):
    name: Literal["O Tides, I Have Returned"] = "O Tides, I Have Returned"
    damage: int = 2
    back_damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status("Sourcewater Droplet", {"usage": 2})
        ]


# Talents


class HeirToTheAncientSeasAuthority_4_5(SkillTalent):
    name: Literal["Heir to the Ancient Sea's Authority"]
    version: Literal["4.5"] = "4.5"
    character_name: Literal["Neuvillette"] = "Neuvillette"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=1, any_dice_number=2
    )
    skill: Literal["As Water Seeks Equilibrium"] = "As Water Seeks Equilibrium"

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        if self.position.not_satisfy(
            "source area=character and both pidx=same and target area=skill",
            target=event.final_damage.position,
        ):
            # not equip, not our use skill
            return []
        if ElementType.HYDRO not in event.reacted_elements:
            # not trigger hydro reaction
            return []
        return [
            CreateObjectAction(
                object_name="Heir to the Ancient Sea's Authority",
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]


# character base


# character class name should contain its version.
class Neuvillette_4_5(CharacterBase):
    name: Literal["Neuvillette"]
    version: Literal["4.5"] = "4.5"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase
        | EquitableJudgment
        | OTearsIShallRepay
        | OTidesIHaveReturned
    ] = []
    faction: List[FactionType] = [FactionType.FONTAINE, FactionType.ARKHE_PNEUMA]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="As Water Seeks Equilibrium",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            EquitableJudgment(),
            OTearsIShallRepay(),
            OTidesIHaveReturned(),
        ]


# define descriptions of newly defined classes. Note key of skills and talents
# have character names. For balance changes, only descs are needed to define.
character_descs: Dict[str, DescDictType] = {
    "CHARACTER/Neuvillette": {
        "names": {"en-US": "Neuvillette", "zh-CN": "那维莱特"},
        "descs": {"4.5": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Avatar_Neuvillette.png",  # noqa: E501
        "id": 1210,
    },
    "CHARACTER_STATUS/Heir to the Ancient Sea's Authority": {
        "names": {"en-US": "Heir to the Ancient Sea's Authority", "zh-CN": "遗龙之荣"},
        "descs": {
            "4.5": {
                "en-US": "This character deals +1 DMG for the next 2 instances.",
                "zh-CN": "角色接下来2次造成的伤害+1。",
            }
        },
    },
    "TALENT_Neuvillette/Heir to the Ancient Sea's Authority": {
        "names": {
            "en-US": "Heir to the Ancient Sea's Authority",
            "zh-CN": "古海孑遗的权柄",
        },
        "descs": {
            "4.5": {
                "en-US": "Combat Action: When your active character is Neuvillette, equip this card.\nAfter Neuvillette equips this card, immediately use As Water Seeks Equilibrium once.\nAfter your characters trigger Hydro-Related Reactions: Neuvillette, who has this card equipped, deals +1 DMG for the next 2 instances.\n(You must have Neuvillette in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为那维莱特时，装备此牌。\n那维莱特装备此牌后，立刻使用一次如水从平。\n我方角色引发水元素相关反应后：装备有此牌的那维莱特接下来2次造成的伤害+1。\n（牌组中包含那维莱特，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Neuvillette.png",  # noqa: E501
        "id": 212101,
    },
    "SKILL_Neuvillette_NORMAL_ATTACK/As Water Seeks Equilibrium": {
        "names": {"en-US": "As Water Seeks Equilibrium", "zh-CN": "如水从平"},
        "descs": {
            "4.5": {"en-US": "Deals 1 Hydro DMG.", "zh-CN": "造成1点水元素伤害。"}
        },
    },
    "SKILL_Neuvillette_ELEMENTAL_SKILL/O Tears, I Shall Repay": {
        "names": {"en-US": "O Tears, I Shall Repay", "zh-CN": "泪水啊，我必偿还"},
        "descs": {
            "4.5": {
                "en-US": "Deals 2 Hydro DMG, creates 1 Sourcewater Droplet.",
                "zh-CN": "造成2点水元素伤害，生成源水之滴。",
            }
        },
    },
    "TEAM_STATUS/Sourcewater Droplet": {
        "names": {"en-US": "Sourcewater Droplet", "zh-CN": "源水之滴"},
        "descs": {
            "4.5": {
                "en-US": "After Neuvillette uses a Normal Attack: Heals the character for 2 HP. After this, Prepare Skill|s1:prepares: Equitable Judgment.\nUsage(s): 1 (Can stack. Max 3 stacks.)",  # noqa: E501
                "zh-CN": "那维莱特进行普通攻击后：治疗角色2点，然后角色准备技能：衡平推裁。\n可用次数：1（可叠加，最多叠加到3次）",  # noqa: E501
            }
        },
        "image_path": "status/Neuvillette_S.png",
    },
    "SKILL_Neuvillette_ELEMENTAL_BURST/O Tides, I Have Returned": {
        "names": {"en-US": "O Tides, I Have Returned", "zh-CN": "潮水啊，我已归来"},
        "descs": {
            "4.5": {
                "en-US": "Deals 2 Hydro DMG, deals 1 Piercing DMG to all opposing characters on standby, then creates a Sourcewater Droplet with 2 Usages.",  # noqa: E501
                "zh-CN": "造成2点水元素伤害，对所有后台敌人造成1点穿透伤害，生成可用次数为2的源水之滴。",  # noqa: E501
            }
        },
    },
    "CHARACTER_STATUS/Equitable Judgment": {
        "names": {"en-US": "Equitable Judgment", "zh-CN": "衡平推裁"},
        "descs": {
            "4.5": {
                "en-US": "(Prepare for 1 turn)\nDeals 2 Hydro DMG. If character has at least 6 HP, then they deal 1 Piercing DMG to themselves and deal +1 DMG.",  # noqa: E501
                "zh-CN": "（需准备1个行动轮）\n造成2点水元素伤害，如果生命值至少为6，则对自身造成1点穿透伤害使伤害+1。",  # noqa: E501
            }
        },
    },
    "SKILL_Neuvillette_NORMAL_ATTACK/Equitable Judgment": {
        "names": {"en-US": "Equitable Judgment", "zh-CN": "衡平推裁"},
        "descs": {
            "4.5": {
                "en-US": "(Prepare for 1 turn)\nDeals 2 Hydro DMG. If character has at least 6 HP, then they deal 1 Piercing DMG to themselves and deal +1 DMG.",  # noqa: E501
                "zh-CN": "（需准备1个行动轮）\n造成2点水元素伤害，如果生命值至少为6，则对自身造成1点穿透伤害使伤害+1。",  # noqa: E501
            }
        },
    },
}


register_class(
    Neuvillette_4_5
    | HeirToTheAncientSeasAuthority_4_5
    | HeirToTheAncientSeasAuthorityStatus_4_5
    | SourcewaterDroplet_4_5
    | EquitableJudgment_4_5,
    character_descs,
)



from typing import Any, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import DamageIncreaseValue

from ....struct import Cost

from ....consts import ElementalReactionType, ObjectPositionType, ObjectType, WeaponType
from .base import RoundEffectWeaponBase


class AThousandFloatingDreams_3_7(RoundEffectWeaponBase):
    name: Literal["A Thousand Floating Dreams"]
    type: Literal[ObjectType.WEAPON] = ObjectType.WEAPON
    version: Literal["3.7"] = "3.7"
    weapon_type: WeaponType = WeaponType.CATALYST

    cost: Cost = Cost(same_dice_number=3)
    max_usage_per_round: int = 2

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        First +1 DMG if self character use skill. Then if this damage is
        our character use skill, and trigger element reaction, +1 DMG.
        """
        # first +1 DMG
        super().value_modifier_DAMAGE_INCREASE(value, match, mode)

        # second element reaction +1 DMG
        if self.usage == 0:
            # no usage left
            return value
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return value
        if value.element_reaction == ElementalReactionType.NONE:
            # no elemental reaction
            return value
        if value.damage_from_element_reaction:
            # from elemental reaction
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self player use skill
            return value
        # modify damage
        assert mode == "REAL"
        value.damage += 1
        self.usage -= 1
        return value


register_class(AThousandFloatingDreams_3_7)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....action import CreateObjectAction, DrawCardAction

from ....struct import Cost
from .base import WeaponBase
from ....consts import ObjectPositionType, WeaponType


class FruitOfFulfillment_3_8(WeaponBase):
    name: Literal["Fruit of Fulfillment"]
    cost: Cost = Cost(any_dice_number=3)
    version: Literal["3.8"] = "3.8"
    weapon_type: WeaponType = WeaponType.CATALYST

    def equip(self, match: Any) -> List[DrawCardAction]:
        """
        draw 2 cards
        """
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=2,
                draw_if_filtered_not_enough=True,
            )
        ]


class KingsSquire_4_0(WeaponBase):
    name: Literal["King's Squire"]
    cost: Cost = Cost(same_dice_number=3)
    version: Literal["4.0"] = "4.0"
    weapon_type: WeaponType = WeaponType.BOW

    def equip(self, match: Any) -> List[CreateObjectAction]:
        """
        attach status
        """
        return [
            CreateObjectAction(
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_name=self.name,
                object_arguments={},
            )
        ]


class Moonpiercer_4_1(KingsSquire_4_0):
    name: Literal["Moonpiercer"]
    version: Literal["4.1"] = "4.1"
    weapon_type: WeaponType = WeaponType.POLEARM


register_class(FruitOfFulfillment_3_8 | KingsSquire_4_0 | Moonpiercer_4_1)



# type: ignore
"""
This file contains a template for writing tests. You can copy this file to
tests/folder/test_xxx.py and modify it to write your own tests.

NOTE: outdated, now better to use `do_log_tests` utility function to test on a log.
"""
from lpsim.agents import InteractionAgent
from lpsim import Deck, Match, MatchState
from tests.utils_for_test import (
    check_hp, get_random_state, get_test_id_from_command, make_respond, 
    set_16_omni
)


def test_...():
    # use frontend and FastAPI server to perform commands, and test commands 
    # that start with TEST. NO NOT put TEST at the end of command list!
    # If a command is successfully performed, frontend will print history 
    # commands in console. Note that frontend cannot distinguish if a new
    # match begins, so you need to refresh the page before recording a new
    # match, otherwise the history commands will be mixed.
    #
    # for tests, it starts with TEST and contains a test id, which is used to
    # identify the test. Other texts in the command are ignored, but you can
    # parse them if you want. Refer to the following code.
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 0 1 2",
            "end",
            "TEST 1 9 10 10 10 9 10",
            ...
        ],
        [
            "sw_card",
            "choose 1",
            "skill 1 0 1 2",
            "end",
            "end",
            "TEST 1 9 10 10 9 8 9",
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        character:Nahida*3
        Rana*30
        ...
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                ...
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_...()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import SkillEndEventArguments

from ...action import Actions, MakeDamageAction, SwitchCharacterAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillBase,
    SkillTalent,
)


# Summons


class ShadowswordLoneGale_3_3(AttackerSummonBase):
    name: Literal["Shadowsword: Lone Gale"]
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.ANEMO
    damage: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If kenki use elemental burst, make damage
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not our player skill made damage, do nothing
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_BURST:
            # not elemental burst, do nothing
            return []
        character = match.player_tables[self.position.player_idx].characters[
            event.action.position.character_idx
        ]
        if character.name != "Maguu Kenki":
            # not kenki, do nothing
            return []
        target_character = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target_character.position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_elemental_type,
                        cost=Cost(),
                    )
                ]
            )
        ]


class ShadowswordGallopingFrost_3_3(ShadowswordLoneGale_3_3):
    name: Literal["Shadowsword: Galloping Frost"]
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1


# Skills


class BlusteringBlade(ElementalSkillBase):
    name: Literal["Blustering Blade"] = "Blustering Blade"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        only create object
        """
        return [self.create_summon("Shadowsword: Lone Gale"), self.charge_self(1)]


class FrostyAssault(ElementalSkillBase):
    name: Literal["Frosty Assault"] = "Frosty Assault"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        only create object
        """
        return [self.create_summon("Shadowsword: Galloping Frost"), self.charge_self(1)]


class PseudoTenguSweeper(ElementalBurstBase):
    name: Literal["Pseudo Tengu Sweeper"] = "Pseudo Tengu Sweeper"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=3
    )


# Talents


class TranscendentAutomaton_3_3(SkillTalent):
    name: Literal["Transcendent Automaton"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Maguu Kenki"] = "Maguu Kenki"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO,
        elemental_dice_number=3,
    )
    skill: Literal["Blustering Blade"] = "Blustering Blade"

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[SwitchCharacterAction]:
        """
        When self use elemental skill, switch to previous or next character
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            source_area=ObjectPositionType.CHARACTER,
            target_area=ObjectPositionType.SKILL,
        ):
            # not equipped on character, or skill source not self
            return []
        if event.action.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill, do nothing
            return []
        table = match.player_tables[self.position.player_idx]
        skill: SkillBase = match.get_object(event.action.position)
        if skill.name == "Blustering Blade":
            # switch to next character
            next_idx = table.next_character_idx()
            if next_idx is None:
                # no next character, do nothing
                return []
            return [
                SwitchCharacterAction(
                    player_idx=self.position.player_idx,
                    character_idx=next_idx,
                )
            ]
        elif skill.name == "Frosty Assault":
            # switch to previous character
            prev_idx = table.previous_character_idx()
            if prev_idx is None:
                # no next character, do nothing
                return []
            return [
                SwitchCharacterAction(
                    player_idx=self.position.player_idx,
                    character_idx=prev_idx,
                )
            ]
        else:
            raise AssertionError("skill name not match")


# character base


class MaguuKenki_3_4(CharacterBase):
    name: Literal["Maguu Kenki"]
    version: Literal["3.4"] = "3.4"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | BlusteringBlade | FrostyAssault | PseudoTenguSweeper
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Ichimonji",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            BlusteringBlade(),
            FrostyAssault(),
            PseudoTenguSweeper(),
        ]


register_class(
    MaguuKenki_3_4
    | TranscendentAutomaton_3_3
    | ShadowswordLoneGale_3_3
    | ShadowswordGallopingFrost_3_3
)



"""
Base classes of objects in the game table. They are all subclasses of
ObjectBase. CardBase are defined here because like ObjectBase, they are used 
in many other places. Other objects are defined in their own files.
"""


import time
import random

from .query import query, query_one

from ..utils.class_registry import register_base_class

from .event import GameStartEventArguments, UseCardEventArguments
from ..utils import BaseModel, accept_same_or_higher_version
from typing import List, Literal, Any, Tuple
from pydantic import validator
from .action import Actions, ActionTypes, CreateObjectAction, RemoveCardAction
from .consts import ObjectType, ObjectPositionType, CostLabels, PlayerActionLabels
from .modifiable_values import CostValue, ModifiableValueTypes
from .struct import DeckRestriction, MultipleObjectPosition, ObjectPosition, Cost


used_object_ids = set()
ID_MOD_NUM = 86400
ID_MULTI_NUM = 1000000
ID_RAND_NUM = 1024


class ObjectBase(BaseModel):
    """
    Base class of objects in the game table. All objects in the game table
    should inherit from this class.
    Args:
        name (str): Name of the object.
        desc (str): Description of the object. When set, frontend
            will add desc to the name, i.e. `{name}_{desc}`, to find
            descriptions. This is useful for objects that have different
            descriptions with different state (talent effect activated etc.).
            Defaults to empty string, and frontend will do nothing.
    """

    name: str
    desc: Literal[""] = ""
    type: ObjectType = ObjectType.EMPTY
    position: ObjectPosition
    id: int = -1

    # If the object is in trashbin, event handlers except those in this list
    # will not available.
    available_handler_in_trashbin: List[ActionTypes] = []

    # If the object is in deck, event handlers except those in this list
    # will not available.
    available_handler_in_deck: List[ActionTypes] = []

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        # check event handler name valid
        for k in dir(self):
            if k[:14] == "event_handler_":
                k = k[14:]
                try:
                    k = ActionTypes(k)
                except ValueError:
                    raise ValueError(f"Invalid event handler name: {k}")
            if k[:15] == "value_modifier_":
                k = k[15:]
                try:
                    k = ModifiableValueTypes(k)
                except ValueError:
                    raise ValueError(f"Invalid value modifier name: {k}")
        # if id is -1, generate a new id
        if self.id == -1:
            self.renew_id()
        used_object_ids.add(self.id)
        # set id in position
        self.position = self.position.set_id(self.id)

    def renew_id(self):
        """
        Renew the id of the object.
        """
        while True:
            new_id = int(
                time.time() % ID_MOD_NUM * ID_MULTI_NUM
            ) * ID_RAND_NUM + random.randint(0, ID_RAND_NUM - 1)
            if new_id in used_object_ids:  # pragma: no cover
                continue
            self.id = new_id
            used_object_ids.add(self.id)
            self.position = self.position.set_id(self.id)
            break

    def query(self, match: Any, command: str) -> List["ObjectBase"]:
        """
        Receives match, and get objects from match based on command.

        Command contains:
        - `both / our / opponent` to select player_table, or `self` to select this
            character (caller area must be character or character_status). must appear
            first.
        - `[active / prev / next]` or
            `[deck / hand / character / team_status / summon / support]`
            for player_table, `[weapon / artifact / talent]` or `[skill / status]` for
            character, to select an object (first group) or object lists (second group).
            Note with `both`, select one object may also receive two objects.
        - `['key=value']` to filter current objects that `x.key == value`. All types
            will be converted to str to compare. it can be used multiple times. command
            can be surrounded with quotes, and if not space in key and value, quotes can
            be omitted.
        - `and` to make multiple queries, and their results are combined.

        Samples:
        - `self` to get self character when called by a character status
        - `opponent active status 'name=Seed of Skandha'` to get status
        - `opponent character status name=Refraction` to get status from all opponent
            characters
        - `both summon` to get all summons on field
        - `self and our active and our next` to select characters
        """
        return query(self.position, match, command)

    def query_one(
        self, match: Any, command: str, allow_multiple: bool = False
    ) -> "ObjectBase | None":
        """
        Receives match, and get one object from match based on command.
        If no object is found, return None. For command structure, see `query` function.
        If allow_multiple is False, when multiple objects are found, raise error.
        """
        return query_one(self.position, match, command, allow_multiple)

    """
    Utility functions that many objects may use and have relaively complex logic, so
    define here and can be called by object.
    TODO: If some logic is restricted to specified base class, put here or under that 
    base class?
    """

    def _check_value_self_skill_or_talent(
        self,
        value: CostValue,
        match: Any,
        skill_cost_label: int = (
            CostLabels.ELEMENTAL_SKILL.value
            | CostLabels.NORMAL_ATTACK.value
            | CostLabels.ELEMENTAL_BURST.value
        ),
    ) -> bool:
        """
        Check whether this object is located on a character, and value is self use
        skill or equip talent to self. This mainly used by artifacts and character
        status.

        Args:
            value (CostValue): The value to check.
            match (Any): The match object.
            skill_cost_label (int): The skill label to check. Defaults to
                accept all three types of skills, i.e. `CostLabels.ELEMENTAL_SKILL.value
                | CostLabels.NORMAL_ATTACK.value | CostLabels.ELEMENTAL_BURST.value`.
        """
        if self.position.area not in [
            ObjectPositionType.CHARACTER,
            ObjectPositionType.SKILL,
            ObjectPositionType.CHARACTER_STATUS,
        ]:
            # self position not right
            return False
        label = value.cost.label
        if label & (skill_cost_label | CostLabels.TALENT.value) == 0:
            # no label match
            return False
        position = value.position
        if self.position.satisfy(
            "both player=same character=same and target area=skill",
            target=position,
        ):
            # self use skill
            return True
        if self.position.satisfy(
            "both player=same and target area=hand", target=position
        ):
            # is hand card, should be talent
            card = match.get_object(position)
            assert card is not None
            character = self.query_one(match, "self")  # self returns character
            assert character is not None
            return card.character_name == character.name
        return False


class CardBase(ObjectBase):
    """
    Base class of all real cards.
    """

    name: str
    type: Literal[
        ObjectType.CARD,
        ObjectType.WEAPON,
        ObjectType.ARTIFACT,
        ObjectType.TALENT,
        ObjectType.SUPPORT,
        ObjectType.ARCANE,
    ] = ObjectType.CARD
    strict_version_validation: bool = False  # default accept higher versions
    version: str
    position: ObjectPosition = ObjectPosition(
        player_idx=-1,
        area=ObjectPositionType.INVALID,
        id=-1,
    )
    cost: Cost
    cost_label: int
    remove_when_used: bool = True
    target_position_type_multiple: Literal[False] = False

    @validator("version", pre=True)
    def accept_same_or_higher_version(cls, v: str, values):
        return accept_same_or_higher_version(cls, v, values)

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        # set cost label into cost
        self.cost.label = self.cost_label
        assert self.cost.original_value is None

    def get_deck_restriction(self) -> DeckRestriction:
        """
        Get the deck restriction of the card. It will be checked when deck is
        created.
        """
        return DeckRestriction(type="NONE", name="", number=0)

    def get_action_type(self, match: Any) -> Tuple[int, bool]:
        """
        Get the action type of using the card.

        Returns:
            Tuple[int, bool]: The first element is the action label, the second
                element is whether the action type is a combat action.
        """
        return PlayerActionLabels.CARD.value, False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Get the targets of the card.
        """
        raise NotImplementedError()

    def get_actions(self, target: ObjectPosition | None, match: Any) -> List[Actions]:
        """
        Act the card. It will return a list of actions.

        Arguments:
            target (ObjectPosition | None): The target of the action.
                for cards that do not need to specify target, target is None.
                Note: the ID of the target may not be the same as the ID of the
                card, e.g. equipping artifact. Reconstruct correct
                ObjectPosition if needed.
            match (Any): The match object.
        """
        raise NotImplementedError()

    def is_valid(self, match: Any) -> bool:
        """
        Check if the card can be used. Note that this function will not check
        the cost of the card.
        """
        return True

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        """
        If this card is used, trigger events
        """
        actions: List[Actions] = []
        if event.action.card_position.id != self.id:
            # not this card
            return actions
        if self.remove_when_used:
            actions.append(
                RemoveCardAction(
                    position=ObjectPosition(
                        player_idx=self.position.player_idx,
                        area=ObjectPositionType.HAND,
                        id=self.id,
                    ),
                    remove_type="USED",
                )
            )
        # target is None, otherwise should match class
        assert event.action.target is None or (
            isinstance(event.action.target, MultipleObjectPosition)
            == self.target_position_type_multiple
        )
        if event.use_card_success:
            # use success, add actions of the card
            actions += self.get_actions(
                target=event.action.target,  # type: ignore
                match=match,
            )
        return actions


class EventCardBase(CardBase):
    type: Literal[ObjectType.CARD] = ObjectType.CARD
    cost_label: int = CostLabels.CARD.value | CostLabels.EVENT.value


class MultiTargetEventCardBase(EventCardBase):
    """
    Base class of cards that can target multiple targets.
    """

    target_position_type_multiple: Literal[True] = True

    def get_targets(self, match: Any) -> List[MultipleObjectPosition]:
        raise NotImplementedError()

    def get_actions(
        self, target: MultipleObjectPosition | None, match: Any
    ) -> List[Actions]:
        """
        Act the card. It will return a list of actions.

        Arguments:
            target (MultipleObjectPosition | None): The target of the action.
                for cards that do not need to specify target, target is None.
                Note: the ID of the target may not be the same as the ID of the
                card, e.g. equipping artifact. Reconstruct correct
                ObjectPosition if needed.
            match (Any): The match object.
        """
        raise NotImplementedError()


class CreateSystemEventHandlerObject(BaseModel):
    """
    objects that will create system event handler at game start
    """

    handler_name: str

    available_handler_in_deck: List[ActionTypes] = [
        ActionTypes.GAME_START,
    ]

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When game start, create event handler with specified name
        """
        return [
            CreateObjectAction(
                object_name=self.handler_name,
                object_position=ObjectPosition(
                    player_idx=-1,
                    area=ObjectPositionType.SYSTEM,
                    id=-1,
                ),
                object_arguments={},
            )
        ]

    def _get_event_handler(self, match: Any):
        """
        get created event handler object. If not found, raise error.
        """
        target_event_handler: Any = None
        for event_handler in match.event_handlers:
            if event_handler.name == self.handler_name:
                target_event_handler = event_handler
                break
        else:
            raise AssertionError(f"event handler {self.handler_name} not found")
        return target_event_handler


# CardBases = CardBase | MultiTargetEventCardBase
register_base_class(CardBase)
register_base_class(EventCardBase)
register_base_class(MultiTargetEventCardBase)



from enum import Enum


class ElementType(str, Enum):
    """
    Enum representing the type of an element.
    """

    NONE = "NONE"
    CRYO = "CRYO"
    HYDRO = "HYDRO"
    PYRO = "PYRO"
    ELECTRO = "ELECTRO"
    GEO = "GEO"
    DENDRO = "DENDRO"
    ANEMO = "ANEMO"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


ELEMENT_DEFAULT_ORDER = [
    ElementType.CRYO,
    ElementType.HYDRO,
    ElementType.PYRO,
    ElementType.ELECTRO,
    ElementType.GEO,
    ElementType.DENDRO,
    ElementType.ANEMO,
]


class DieColor(str, Enum):
    """
    Enum representing the color of a die. It can also be called as elemental
    type, elemental attributes. Besides existing element types, there are
    also Onmi, which can be used as any color.

    Attributes:
        CRYO (str): The die color is cryo.
        HYDRO (str): The die color is hydro.
        PYRO (str): The die color is pyro.
        ELECTRO (str): The die color is electro.
        GEO (str): The die color is geo.
        DENDRO (str): The die color is dendro.
        ANEMO (str): The die color is anemo.
        OMNI (str): The die color is omni.
    """

    CRYO = "CRYO"
    HYDRO = "HYDRO"
    PYRO = "PYRO"
    ELECTRO = "ELECTRO"
    GEO = "GEO"
    DENDRO = "DENDRO"
    ANEMO = "ANEMO"
    OMNI = "OMNI"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


ELEMENT_TO_DIE_COLOR = {
    ElementType.CRYO: DieColor.CRYO,
    ElementType.HYDRO: DieColor.HYDRO,
    ElementType.PYRO: DieColor.PYRO,
    ElementType.ELECTRO: DieColor.ELECTRO,
    ElementType.GEO: DieColor.GEO,
    ElementType.DENDRO: DieColor.DENDRO,
    ElementType.ANEMO: DieColor.ANEMO,
}
DIE_COLOR_TO_ELEMENT = {
    DieColor.CRYO: ElementType.CRYO,
    DieColor.HYDRO: ElementType.HYDRO,
    DieColor.PYRO: ElementType.PYRO,
    DieColor.ELECTRO: ElementType.ELECTRO,
    DieColor.GEO: ElementType.GEO,
    DieColor.DENDRO: ElementType.DENDRO,
    DieColor.ANEMO: ElementType.ANEMO,
}


class DamageElementalType(str, Enum):
    """
    Enum representing the type of a damage.
    """

    # elemental damage
    CRYO = "CRYO"
    HYDRO = "HYDRO"
    PYRO = "PYRO"
    ELECTRO = "ELECTRO"
    GEO = "GEO"
    DENDRO = "DENDRO"
    ANEMO = "ANEMO"

    # other damage
    PHYSICAL = "PHYSICAL"
    PIERCING = "PIERCING"
    HEAL = "HEAL"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


ELEMENT_TO_DAMAGE_TYPE = {
    ElementType.NONE: DamageElementalType.PHYSICAL,
    ElementType.CRYO: DamageElementalType.CRYO,
    ElementType.HYDRO: DamageElementalType.HYDRO,
    ElementType.PYRO: DamageElementalType.PYRO,
    ElementType.ELECTRO: DamageElementalType.ELECTRO,
    ElementType.GEO: DamageElementalType.GEO,
    ElementType.DENDRO: DamageElementalType.DENDRO,
    ElementType.ANEMO: DamageElementalType.ANEMO,
}
DAMAGE_TYPE_TO_ELEMENT = {
    DamageElementalType.CRYO: ElementType.CRYO,
    DamageElementalType.HYDRO: ElementType.HYDRO,
    DamageElementalType.PYRO: ElementType.PYRO,
    DamageElementalType.ELECTRO: ElementType.ELECTRO,
    DamageElementalType.GEO: ElementType.GEO,
    DamageElementalType.DENDRO: ElementType.DENDRO,
    DamageElementalType.ANEMO: ElementType.ANEMO,
    DamageElementalType.PHYSICAL: ElementType.NONE,
    DamageElementalType.PIERCING: ElementType.NONE,
    DamageElementalType.HEAL: ElementType.NONE,
}


class DamageType(str, Enum):
    """
    Enum representing the type of a damage.
    """

    DAMAGE = "DAMAGE"
    HEAL = "HEAL"
    ELEMENT_APPLICATION = "ELEMENT_APPLICATION"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class ObjectType(str, Enum):
    """
    Enum representing the type of an object.
    """

    EMPTY = "EMPTY"
    CHARACTER = "CHARACTER"
    DICE = "DICE"
    CARD = "CARD"
    SUMMON = "SUMMON"
    SUPPORT = "SUPPORT"
    SKILL = "SKILL"
    WEAPON = "WEAPON"
    ARTIFACT = "ARTIFACT"
    TALENT = "TALENT"
    CHARACTER_STATUS = "CHARACTER_STATUS"
    TEAM_STATUS = "TEAM_STATUS"
    ARCANE = "ARCANE"
    SYSTEM = "SYSTEM"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class ObjectPositionType(str, Enum):
    """
    Enum representing the position of an object.
    """

    DECK = "DECK"
    HAND = "HAND"
    SUMMON = "SUMMON"
    SUPPORT = "SUPPORT"
    DICE = "DICE"
    CHARACTER = "CHARACTER"
    CHARACTER_STATUS = "CHARACTER_STATUS"
    SKILL = "SKILL"
    TEAM_STATUS = "TEAM_STATUS"
    SYSTEM = "SYSTEM"
    INVALID = "INVALID"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class FactionType(str, Enum):
    """
    Factions of characters.
    """

    MONDSTADT = "MONDSTADT"
    LIYUE = "LIYUE"
    INAZUMA = "INAZUMA"
    SUMERU = "SUMERU"
    FONTAINE = "FONTAINE"
    NATLAN = "NATLAN"
    SNEZHNAYA = "SNEZHNAYA"

    FATUI = "FATUI"
    MONSTER = "MONSTER"
    HILICHURL = "HILICHURL"

    ARKHE_PNEUMA = "ARKHE_PNEUMA"
    ARKHE_OUSIA = "ARKHE_OUSIA"

    THE_EREMITES = "THE_EREMITES"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class WeaponType(str, Enum):
    """
    Types of weapons.
    """

    SWORD = "SWORD"
    CLAYMORE = "CLAYMORE"
    POLEARM = "POLEARM"
    CATALYST = "CATALYST"
    BOW = "BOW"
    OTHER = "OTHER"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class SkillType(str, Enum):
    """
    Types of skills.
    """

    NORMAL_ATTACK = "NORMAL_ATTACK"
    ELEMENTAL_SKILL = "ELEMENTAL_SKILL"
    ELEMENTAL_BURST = "ELEMENTAL_BURST"
    PASSIVE = "PASSIVE"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class ElementalReactionType(str, Enum):
    """
    Types of elemental reactions.
    """

    NONE = "NONE"
    MELT = "MELT"
    VAPORIZE = "VAPORIZE"
    OVERLOADED = "OVERLOADED"
    SUPERCONDUCT = "SUPERCONDUCT"
    ELECTROCHARGED = "ELECTROCHARGED"
    FROZEN = "FROZEN"
    SWIRL = "SWIRL"
    CRYSTALLIZE = "CRYSTALLIZE"
    BURNING = "BURNING"
    BLOOM = "BLOOM"
    QUICKEN = "QUICKEN"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class CostLabels(int, Enum):
    SWITCH_CHARACTER = 0x1
    CARD = 0x2
    COMPANION = 0x4
    LOCATION = 0x8
    ITEM = 0x10
    ARTIFACT = 0x20
    WEAPON = 0x40
    TALENT = 0x80
    FOOD = 0x100
    ARCANE = 0x200
    NORMAL_ATTACK = 0x400
    ELEMENTAL_SKILL = 0x800
    ELEMENTAL_BURST = 0x1000
    CHARGED_ATTACK = 0x2000
    PLUNGING_ATTACK = 0x4000
    # equipment is all artifact, all weapon and some talents
    EQUIPMENT = 0x8000
    # event is some action cards, and some talents
    EVENT = 0x10000

    SKILLS = 0x1C00
    SUPPORTS = 0x1C

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class IconType(str, Enum):
    """
    Icon types of objects. Used for display. All summons, status, and supports
    should have one of these types.
    """

    # e.g. ???
    BUFF = "BUFF"
    # e.g. I Haven't lost yet used
    DEBUFF = "DEBUFF"
    # e.g. Changing Shifts
    SPECIAL = "SPECIAL"
    # e.g. enemy hydro abyss mage
    DAZE = "DAZE"
    # e.g. enemy dog
    DOT = "DOT"
    # Frozen
    FROZEN = "FROZEN"

    # e.g. Collei talent
    ATK_SELF = "ATK_SELF"
    # e.g. Tandoori Chicken
    ATK_UP = "ATK_UP"
    # e.g. kazuha talent
    ATK_UP_FIRE = "ELEMENT_ATK_UP_FIRE"
    ATK_UP_ELEC = "ELEMENT_ATK_UP_ELEC"
    ATK_UP_ICE = "ELEMENT_ATK_UP_ICE"
    ATK_UP_WATER = "ELEMENT_ATK_UP_WATER"
    ATK_UP_WIND = "ELEMENT_ATK_UP_WIND"
    ATK_UP_ROCK = "ELEMENT_ATK_UP_ROCK"
    ATK_UP_GRASS = "ELEMENT_ATK_UP_GRASS"
    # e.g. keqing, ayaka
    ELEMENT_ENCHANT_FIRE = "ELEMENT_ENCHANT_FIRE"
    ELEMENT_ENCHANT_ELEC = "ELEMENT_ENCHANT_ELEC"
    ELEMENT_ENCHANT_ICE = "ELEMENT_ENCHANT_ICE"
    ELEMENT_ENCHANT_WATER = "ELEMENT_ENCHANT_WATER"
    ELEMENT_ENCHANT_WIND = "ELEMENT_ENCHANT_WIND"
    ELEMENT_ENCHANT_ROCK = "ELEMENT_ENCHANT_ROCK"
    ELEMENT_ENCHANT_GRASS = "ELEMENT_ENCHANT_GRASS"
    # debuffs
    DEBUFF_ELEMENT_FIRE = "DEBUFF_ELEMENT_FIRE"
    DEBUFF_ELEMENT_ELEC = "DEBUFF_ELEMENT_ELEC"
    DEBUFF_ELEMENT_ICE = "DEBUFF_ELEMENT_ICE"
    DEBUFF_ELEMENT_WATER = "DEBUFF_ELEMENT_WATER"
    DEBUFF_ELEMENT_WIND = "DEBUFF_ELEMENT_WIND"
    DEBUFF_ELEMENT_ROCK = "DEBUFF_ELEMENT_ROCK"
    DEBUFF_ELEMENT_GRASS = "DEBUFF_ELEMENT_GRASS"
    DEBUFF_ELEMENT_PHYSICS = "DEBUFF_ELEMENT_PHYSICS"
    # e.g. Lotus Flower Crisp, Ushi
    BARRIER = "BARRIER"
    # e.g. Crystallize, Unmovable Mountain
    SHIELD = "SHIELD"
    # e.g. pizza
    HEAL = "HEAL"
    # e.g. Electro Hypostasis
    REVIVE = "REVIVE"

    # food card used for character, or egg used for team.
    FOOD = "FOOD"

    # e.g. most of summons
    TIMESTATE = "TIMESTATE"
    # e.g. Timmie, Liben
    COUNTER = "COUNTER"

    # e.g. Jade Chamber, Traveler's Handy Sword
    NONE = "NONE"

    # with others, the status has its special icon based on its name
    OTHERS = "OTHERS"

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


ELEMENT_TO_ATK_UP_ICON = {
    ElementType.CRYO: IconType.ATK_UP_ICE,
    ElementType.HYDRO: IconType.ATK_UP_WATER,
    ElementType.PYRO: IconType.ATK_UP_FIRE,
    ElementType.ELECTRO: IconType.ATK_UP_ELEC,
    ElementType.GEO: IconType.ATK_UP_ROCK,
    ElementType.DENDRO: IconType.ATK_UP_GRASS,
    ElementType.ANEMO: IconType.ATK_UP_WIND,
}


ELEMENT_TO_ENCHANT_ICON = {
    ElementType.CRYO: IconType.ELEMENT_ENCHANT_ICE,
    ElementType.HYDRO: IconType.ELEMENT_ENCHANT_WATER,
    ElementType.PYRO: IconType.ELEMENT_ENCHANT_FIRE,
    ElementType.ELECTRO: IconType.ELEMENT_ENCHANT_ELEC,
    ElementType.GEO: IconType.ELEMENT_ENCHANT_ROCK,
    ElementType.DENDRO: IconType.ELEMENT_ENCHANT_GRASS,
    ElementType.ANEMO: IconType.ELEMENT_ENCHANT_WIND,
}


class PlayerActionLabels(int, Enum):
    SKILL = 0x1
    SWITCH = 0x2
    CARD = 0x4
    TUNE = 0x8
    END = 0x10

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name



from typing import Dict, List, Literal

from lpsim.server.action import CreateObjectAction
from lpsim.server.card.equipment.weapon.base import RoundEffectWeaponBase
from lpsim.server.consts import DamageType, IconType, ObjectPositionType, WeaponType
from lpsim.server.event import ReceiveDamageEventArguments
from lpsim.server.match import Match
from lpsim.server.modifiable_values import DamageIncreaseValue
from lpsim.server.status.character_status.base import (
    RoundCharacterStatus,
    UsageCharacterStatus,
)
from lpsim.server.struct import Cost
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class TomeOfTheEternalFlowStatus_4_5(UsageCharacterStatus, RoundCharacterStatus):
    name: Literal["Tome of the Eternal Flow"] = "Tome of the Eternal Flow"
    version: Literal["4.5"] = "4.5"
    icon_type: IconType = IconType.ATK_UP  # TODO
    usage: int = 1
    max_usage: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        assert mode == "REAL"
        if self.usage <= 0:  # pragma: no cover
            # no usage, not modify
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not this character use skill, not modify
            return value
        # modify
        value.damage += 2
        self.usage -= 1
        return value


class TomeOfTheEternalFlow_4_5(RoundEffectWeaponBase):
    name: Literal["Tome of the Eternal Flow"] = "Tome of the Eternal Flow"
    version: Literal["4.5"] = "4.5"
    cost: Cost = Cost(same_dice_number=3)
    weapon_type: WeaponType = WeaponType.CATALYST
    # we use round effect to reset. when usage to 2, then create a status.
    max_usage_per_round: int = 0

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        if self.position.not_satisfy(
            "both pidx=same cidx=same", event.final_damage.target_position
        ):
            return []
        if event.final_damage.damage_type == DamageType.ELEMENT_APPLICATION:
            return []
        self.usage += 1
        if self.usage == 2:
            return [
                CreateObjectAction(
                    object_name="Tome of the Eternal Flow",
                    object_position=self.position.set_area(
                        ObjectPositionType.CHARACTER_STATUS
                    ),
                    object_arguments={},
                )
            ]
        return []


desc: Dict[str, DescDictType] = {
    "CHARACTER_STATUS/Tome of the Eternal Flow": {
        "names": {"en-US": "Tome of the Eternal Flow", "zh-CN": "万世流涌大典"},
        "descs": {
            "4.5": {
                "en-US": "This character's next attack this Round deals +2 additional DMG.",  # noqa: E501
                "zh-CN": "角色本回合中下次造成的伤害+2。",
            }
        },
    },
    "WEAPON/Tome of the Eternal Flow": {
        "names": {"en-US": "Tome of the Eternal Flow", "zh-CN": "万世流涌大典"},
        "descs": {
            "4.5": {
                "en-US": "The character deals +1 DMG.\nAfter this character takes DMG or is healed: If this character has already been damaged or healed 2 times this Round, then this character's next attack this Round deals +2 additional DMG. (Once per Round)\n(Only Catalyst Characters can equip this. A character can equip a maximum of 1 Weapon)",  # noqa: E501
                "zh-CN": "角色造成的伤害+1。\n角色受到伤害或治疗后：如果本回合已受到伤害或治疗累计2次，则角色本回合中下次造成的伤害+2。（每回合1次）\n（「法器」角色才能装备。角色最多装备1件「武器」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Weapon_Wanshiliuyongdadian.png",  # noqa: E501
        "id": 311108,
    },
}


register_class(TomeOfTheEternalFlow_4_5 | TomeOfTheEternalFlowStatus_4_5, desc)



from typing import Dict, Literal

import pytest
from lpsim.server.card.support.companions import CompanionBase
from lpsim.server.card.event.foods import FoodCardBase
from lpsim.server.card.support.base import SupportBase
from lpsim.server.consts import DieColor
from lpsim.server.summon.base import SummonBase
from lpsim.server.character.character_base import CharacterBase
from lpsim.utils.desc_registry import (
    DescDictType,
    desc_exist,
    get_desc_patch,
    update_cost,
    update_desc,
)
from lpsim.utils.class_registry import (
    get_class_list_by_base_class,
    get_instance,
    register_class,
)
from lpsim.server.struct import Cost
from lpsim.server.object_base import EventCardBase


def test_class_registry():
    class Sleep_1_0(EventCardBase):
        name: Literal["Sleep"]
        version: Literal["1.0"] = "1.0"
        cost: Cost = Cost()

    sleep_desc: Dict[str, DescDictType] = {
        "CARD/Sleep": {
            "names": {
                "zh-CN": "睡觉",
                "en-US": "Sleep",
            }
        }
    }

    with pytest.raises(AssertionError):
        # no desc, cannot register
        register_class(Sleep_1_0)
    with pytest.raises(AssertionError):
        # desc without version info, cannot register
        register_class(Sleep_1_0, sleep_desc)
    sleep_desc["CARD/Sleep"]["descs"] = {
        "3.0": {
            "zh-CN": "睡觉",
            "en-US": "Sleep",
        }
    }
    with pytest.raises(AssertionError):
        # desc with wrong version info, cannot register
        register_class(Sleep_1_0, sleep_desc)
    sleep_desc["CARD/Sleep"]["descs"] = {
        "1.0": {
            "zh-CN": "睡觉",
            "en-US": "Sleep",
        }
    }
    register_class(Sleep_1_0, sleep_desc)
    # duplicate, should print warning
    register_class(Sleep_1_0)
    with pytest.raises(AssertionError):
        # duplicate and not same, raise error
        class Sleep2_1_0(EventCardBase):
            name: Literal["Sleep"]
            version: Literal["1.0"] = "1.0"
            cost: Cost = Cost()

        register_class(Sleep2_1_0)
    with pytest.raises(AssertionError):
        # wrong base class, get error
        get_instance(CharacterBase, {"name": "Sleep", "version": "1.0"})
    with pytest.raises(AssertionError):
        # wrong base classes, get error
        get_instance(CharacterBase | SummonBase, {"name": "Sleep", "version": "1.0"})
    # contain right class, get instance
    ins = get_instance(EventCardBase, {"name": "Sleep", "version": "1.0"})
    assert ins.name == "Sleep"
    ins = get_instance(SupportBase | EventCardBase, {"name": "Sleep", "version": "1.0"})
    assert ins.name == "Sleep"
    ins = get_instance(EventCardBase, {"name": "Sleep"})
    assert ins.name == "Sleep"
    with pytest.raises(AssertionError):
        # wrong version, get error
        get_instance(EventCardBase, {"name": "Sleep", "version": "0.1"})


def test_desc_registry():
    sleep_desc: Dict[str, DescDictType] = {
        "CARD/SleepDesc": {
            "names": {
                "zh-CN": "睡觉",
                "en-US": "Sleep",
            },
            "id": 999998,
            "image_path": "http://www.baidu.com",
            "descs": {
                "1.0": {
                    "zh-CN": "睡觉",
                    "en-US": "Sleep",
                }
            },
        }
    }
    assert not desc_exist("CARD", "SleepDesc", "1.0")
    update_desc(sleep_desc)
    assert desc_exist("CARD", "SleepDesc", "1.0")
    assert not desc_exist("CARD", "SleepDesc", "2.0")
    # can add again, but print warning
    update_desc(sleep_desc)
    sleep_desc["CARD/SleepDesc"]["descs"]["1.0"]["zh-CN"] = "睡觉1"
    with pytest.raises(ValueError):
        # duplicate key, raise error
        update_desc(sleep_desc)
    sleep_desc["CARD/SleepDesc"]["descs"]["1.0"]["zh-CN"] = "睡觉"
    sleep_desc["CARD/SleepDesc"]["names"]["zh-CN"] = "睡觉1"
    with pytest.raises(ValueError):
        # duplicate key, raise error
        update_desc(sleep_desc)
    sleep_patch: Dict[str, DescDictType] = {
        "CARD/SleepDesc": {
            "descs": {
                "2.0": {
                    "zh-CN": "睡觉2",
                    "en-US": "Sleep2",
                }
            }
        }
    }
    # only update new version descs
    update_desc(sleep_patch)
    assert desc_exist("CARD", "SleepDesc", "2.0")
    sleep_patch_2: Dict[str, DescDictType] = {
        "CARD/SleepDesc": {
            "descs": {
                "3.0": {
                    "zh-CN": "$CARD/SleepDesc|descs|1.0|zh-CN",
                    "en-US": "$CARD/SleepDesc|descs|1.0|en-US",
                }
            }
        }
    }
    update_desc(sleep_patch_2)
    # use key reference to update
    patch = get_desc_patch()
    assert "CARD/SleepDesc" in patch
    assert "descs" in patch["CARD/SleepDesc"]
    assert patch["CARD/SleepDesc"]["descs"]["3.0"]["zh-CN"] == "睡觉"
    assert patch["CARD/SleepDesc"]["descs"]["2.0"]["zh-CN"] == "睡觉2"
    assert patch["CARD/SleepDesc"]["descs"]["1.0"]["zh-CN"] == "睡觉"
    assert patch["CARD/SleepDesc"]["descs"]["3.0"]["en-US"] == "Sleep"
    assert patch["CARD/SleepDesc"]["descs"]["2.0"]["en-US"] == "Sleep2"
    assert patch["CARD/SleepDesc"]["descs"]["1.0"]["en-US"] == "Sleep"
    sleep_patch_3: Dict[str, DescDictType] = {
        "CARD/SleepDesc": {
            "descs": {
                "3.0": {
                    "zh-CN": "$CARD/SleepDesc|descs|1.0|zh-CN",
                    "en-US": "$CARD/SleepDesc|descs|1.0|en-US",
                }
            }
        }
    }
    # have duplicate key, print warning
    update_desc(sleep_patch_3)
    sleep_patch_4: Dict[str, DescDictType] = {
        "CARD/SleepDesc": {
            "descs": {
                "5.0": {
                    "zh-CN": "$CARD/SleepDesc|descs|4.0|zh-CN",
                    "en-US": "$CARD/SleepDesc|descs|4.0|en-US",
                }
            }
        }
    }
    with pytest.raises(ValueError):
        # wrong key, raise error
        update_desc(sleep_patch_4)
    empty_name: Dict[str, DescDictType] = {
        "CARD/EmptyName": {
            "names": {
                "zh-CN": "",
            }
        }
    }
    with pytest.raises(ValueError):
        # empty name, raise error
        update_desc(empty_name)
    empty_name_2: Dict[str, DescDictType] = {
        "CARD/EmptyName2": {
            "names": {
                "zh-CN": "XXX",
            }
        }
    }
    with pytest.raises(ValueError):
        # empty name, raise error
        update_desc(empty_name_2)


def test_get_class_list():
    nlist = get_class_list_by_base_class(SupportBase, "3.3", set(["NRE"]))
    assert "NRE" not in nlist
    assert "Wangshu Inn" in nlist
    assert len(nlist) == 19  # initial 20 and except NRE
    nlist = get_class_list_by_base_class(SupportBase, "3.3")
    assert "NRE" in nlist
    assert len(nlist) == 20  # initial 20
    nlist = get_class_list_by_base_class(SupportBase | FoodCardBase, "3.3")
    assert len(nlist) == 28  # initial 20 + 8 food cards
    nlist = get_class_list_by_base_class(FoodCardBase, "4.2")
    assert len(nlist) == 12  # in 4.2, there are 12 food cards
    nlist = get_class_list_by_base_class(FoodCardBase, "3.3", set(["NRE"]))
    assert len(nlist) == 8
    nlist = get_class_list_by_base_class(CompanionBase, "3.3")
    assert len(nlist) == 12


def test_existing_cost():
    """
    request existing cost for mobs, Starsigns, and Strategize.
    """
    pyro_skill_cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)
    hydro_normal_cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=1, any_dice_number=2
    )
    dendro_burst_cost = Cost(
        elemental_dice_color=DieColor.DENDRO,
        elemental_dice_number=3,
        charge=2,
    )
    starsigns_cost = Cost(any_dice_number=2)
    strategize_cost = Cost(same_dice_number=1)
    p = get_desc_patch()
    assert "cost" in p["SKILL_PyroMob_ELEMENTAL_SKILL/Pyro Elemental Skill"]
    desc_cost = Cost(
        **p["SKILL_PyroMob_ELEMENTAL_SKILL/Pyro Elemental Skill"]["cost"]["1.0"]
    )
    desc_cost.label = 0
    assert pyro_skill_cost == desc_cost
    assert "cost" in p["SKILL_HydroMobMage_NORMAL_ATTACK/Hydro Normal Attack"]
    desc_cost = Cost(
        **p["SKILL_HydroMobMage_NORMAL_ATTACK/Hydro Normal Attack"]["cost"]["1.0"]
    )
    desc_cost.label = 0
    assert hydro_normal_cost == desc_cost
    assert "cost" in p["SKILL_HydroMob_NORMAL_ATTACK/Physical Normal Attack"]
    desc_cost = Cost(
        **p["SKILL_HydroMob_NORMAL_ATTACK/Physical Normal Attack"]["cost"]["1.0"]
    )
    desc_cost.label = 0
    assert hydro_normal_cost == desc_cost
    assert "cost" in p["SKILL_DendroMob_ELEMENTAL_BURST/Dendro Elemental Burst"]
    desc_cost = Cost(
        **p["SKILL_DendroMob_ELEMENTAL_BURST/Dendro Elemental Burst"]["cost"]["1.0"]
    )
    desc_cost.label = 0
    assert dendro_burst_cost == desc_cost
    assert "cost" in p["CARD/Starsigns"]
    desc_cost = Cost(**p["CARD/Starsigns"]["cost"]["3.3"])
    desc_cost.label = 0
    assert starsigns_cost == desc_cost
    assert "cost" in p["CARD/Strategize"]
    desc_cost = Cost(**p["CARD/Strategize"]["cost"]["3.3"])
    desc_cost.label = 0
    assert strategize_cost == desc_cost


def test_register_cost():
    dd_desc: Dict[str, DescDictType] = {
        "CARD/DD": {
            "names": {
                "zh-CN": "滴滴",
                "en-US": "DD",
            },
            "id": 999998,
            "image_path": "http://www.baidu.com",
            "descs": {
                "1.0": {
                    "zh-CN": "滴滴",
                    "en-US": "DD",
                }
            },
        }
    }
    dd_cost = Cost(same_dice_number=1)
    with pytest.raises(ValueError):
        # no desc, cannot register cost
        update_cost("CARD", "DD", "1.0", dd_cost)
    update_desc(dd_desc)
    p = get_desc_patch()
    assert "CARD/DD" in p
    assert "cost" not in p["CARD/DD"]
    update_cost("CARD", "DD", "1.0", dd_cost)
    p = get_desc_patch()
    assert "CARD/DD" in p
    assert "cost" in p["CARD/DD"]
    assert "1.0" in p["CARD/DD"]["cost"]
    assert p["CARD/DD"]["cost"]["1.0"]["same_dice_number"] == 1
    assert p["CARD/DD"]["cost"]["1.0"]["any_dice_number"] == 0


def test_wrong_id():
    wrong_prefix = ["CHARACTER_STATUS", "SKILL", "SUMMON", "TEAM_STATUS"]
    for wp in wrong_prefix:
        wrong_id_desc: Dict[str, DescDictType] = {
            f"{wp}/DD": {
                "names": {
                    "zh-CN": "滴滴",
                    "en-US": "DD",
                },
                "id": 999998,
                "image_path": "http://www.baidu.com",
                "descs": {
                    "1.0": {
                        "zh-CN": "滴滴",
                        "en-US": "DD",
                    }
                },
            }
        }
        with pytest.raises(ValueError):
            update_desc(wrong_id_desc)
    for wp in wrong_prefix:
        right_desc: Dict[str, DescDictType] = {
            f"{wp}/DDD": {
                "names": {
                    "zh-CN": "滴滴",
                    "en-US": "DD",
                },
                "image_path": "http://www.baidu.com",
                "descs": {
                    "1.0": {
                        "zh-CN": "滴滴",
                        "en-US": "DD",
                    }
                },
            }
        }
        update_desc(right_desc)


if __name__ == "__main__":
    test_class_registry()
    test_desc_registry()
    test_get_class_list()
    test_existing_cost()
    test_register_cost()
    test_wrong_id()



from typing import Any, Dict, List, Literal

from pydantic import PrivateAttr

from lpsim.server.consts import IconType, SkillType
from lpsim.server.status.character_status.base import RoundEndAttackCharacterStatus
from lpsim.utils.class_registry import register_class
from lpsim.server.match import Match
from lpsim.utils.desc_registry import DescDictType
from lpsim.server.summon.base import AttackerSummonBase
from lpsim.server.modifiable_values import DamageValue
from lpsim.server.event import (
    RoundEndEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from lpsim.server.action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    MakeDamageAction,
)
from lpsim.server.struct import Cost, ObjectPosition
from lpsim.consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from lpsim.server.character.character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Character status.


# Round status, will last for several rounds and disappear
class SnappySilhoete_4_5(RoundEndAttackCharacterStatus):
    name: Literal["Snappy Silhouette"] = "Snappy Silhouette"
    version: Literal["4.5"] = "4.5"
    usage: int = 2
    max_usage: int = 2
    icon_type: IconType = IconType.DEBUFF_ELEMENT_ICE
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        # when only 1 usage left, and self character has cryo app, increase damage to 2
        who: CharacterBase = self.query_one(match, "self")  # type: ignore
        assert who is not None
        if ElementType.CRYO in who.element_application and self.usage == 1:
            self.damage = 2
        return super().event_handler_ROUND_END(event, match)


# Summons


class NewsflashField_4_5(AttackerSummonBase):
    name: Literal["Newsflash Field"] = "Newsflash Field"
    version: Literal["4.5"] = "4.5"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        heal_targets = self.query(match, "our character is_alive=true")
        for target in heal_targets:
            damage_action.damage_value_list.append(
                DamageValue.create_heal(self.position, target.position, -1, Cost())
            )
        return ret


# Skills


class FramingFreezingPointComposition(ElementalSkillBase):
    name: Literal[
        "Framing: Freezing Point Composition"
    ] = "Framing: Freezing Point Composition"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_opposite_character_status(match, "Snappy Silhouette")
        ]


class StillPhotoComprehensiveConfirmation(ElementalBurstBase):
    name: Literal[
        "Still Photo: Comprehensive Confirmation"
    ] = "Still Photo: Comprehensive Confirmation"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        ret: List[Actions] = super().get_actions(match)
        damage_action = ret[1]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        characters = self.query(match, "our character is_alive=true")
        assert len(characters) > 0
        for character in characters:
            damage_action.damage_value_list.append(
                DamageValue.create_heal(
                    self.position, character.position, -1, self.cost
                )
            )
        ret.append(self.create_summon("Newsflash Field"))
        return ret


# Talents


class ASummationOfInterest_4_5(SkillTalent):
    name: Literal["A Summation of Interest"]
    version: Literal["4.5"] = "4.5"
    character_name: Literal["Charlotte"] = "Charlotte"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)
    skill: Literal[
        "Framing: Freezing Point Composition"
    ] = "Framing: Freezing Point Composition"
    usage: int = 1
    _max_usage: int = PrivateAttr(1)

    def get_actions(self, target: ObjectPosition | None, match: Any) -> List[Actions]:
        self.usage = self._max_usage
        return super().get_actions(target, match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.usage = self._max_usage
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[MakeDamageAction]:
        if self.usage <= 0:
            return []
        if self.position.not_satisfy(
            "source area=character and both pidx=same and target area=skill",
            target=event.action.position,
            match=match,
        ):
            # not equip, or not self use skill
            return []
        if event.action.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        target_status = self.query_one(
            match, 'opponent character status name="Snappy Silhouette"'
        )
        if target_status is None:
            # no target status
            return []
        self.usage -= 1
        target = self.query_one(match, "our active")
        assert target is not None
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue.create_heal(self.position, target.position, -2, Cost())
                ]
            )
        ]


# character base


# character class name should contain its version.
class Charlotte_4_5(CharacterBase):
    name: Literal["Charlotte"]
    version: Literal["4.5"] = "4.5"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase
        | FramingFreezingPointComposition
        | StillPhotoComprehensiveConfirmation
    ] = []
    faction: List[FactionType] = [FactionType.FONTAINE, FactionType.ARKHE_PNEUMA]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Cool-Color Capture",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            FramingFreezingPointComposition(),
            StillPhotoComprehensiveConfirmation(),
        ]


# define descriptions of newly defined classes. Note key of skills and talents
# have character names. For balance changes, only descs are needed to define.
character_descs: Dict[str, DescDictType] = {
    "CHARACTER/Charlotte": {
        "names": {"en-US": "Charlotte", "zh-CN": "夏洛蒂"},
        "descs": {"4.5": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Avatar_Charlotte.png",  # noqa: E501
        "id": 1110,
    },
    "SKILL_Charlotte_NORMAL_ATTACK/Cool-Color Capture": {
        "names": {"en-US": "Cool-Color Capture", "zh-CN": "冷色摄影律"},
        "descs": {
            "4.5": {"en-US": "Deals 1 Cryo DMG.", "zh-CN": "造成1点冰元素伤害。"}
        },
    },
    "SKILL_Charlotte_ELEMENTAL_SKILL/Framing: Freezing Point Composition": {
        "names": {
            "en-US": "Framing: Freezing Point Composition",
            "zh-CN": "取景·冰点构图法",
        },
        "descs": {
            "4.5": {
                "en-US": "Deals 1 Cryo DMG. Attach Snappy Silhouette to the target.",
                "zh-CN": "造成1点冰元素伤害，目标附属瞬时剪影。",
            }
        },
    },
    "CHARACTER_STATUS/Snappy Silhouette": {
        "names": {"en-US": "Snappy Silhouette", "zh-CN": "瞬时剪影"},
        "descs": {
            "4.5": {
                "en-US": "End Phase: Deal 1 Cryo DMG to the character(s) to which this is attached. If only 1 Usage(s) is remaining and the attached character is affected by Cryo, this instance of DMG +1.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "结束阶段：对所附属角色造成1点冰元素伤害；如果可用次数仅剩余1且所附属角色具有冰元素附着，则此伤害+1。\n可用次数：2",  # noqa: E501
            }
        },
    },
    "SKILL_Charlotte_ELEMENTAL_BURST/Still Photo: Comprehensive Confirmation": {
        "names": {
            "en-US": "Still Photo: Comprehensive Confirmation",
            "zh-CN": "定格·全方位确证",
        },
        "descs": {
            "4.5": {
                "en-US": "Deals 1 Cryo DMG, heals all your characters for 1 HP, summons 1 Newsflash Field.",  # noqa: E501
                "zh-CN": "造成1点冰元素伤害，治疗我方所有角色1点，召唤临事场域。",
            }
        },
    },
    "SUMMON/Newsflash Field": {
        "names": {"en-US": "Newsflash Field", "zh-CN": "临事场域"},
        "descs": {
            "4.5": {
                "en-US": "End Phase: Deal 1 Cryo DMG, heal your active character for 1 HP.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "结束阶段：造成1点冰元素伤害，治疗我方出战角色1点。\n可用次数：2",
            }
        },
        "image_path": "cardface/Summon_Charlotte.png",
    },
    "TALENT_Charlotte/A Summation of Interest": {
        "names": {"en-US": "A Summation of Interest", "zh-CN": "以有趣相关为要义"},
        "descs": {
            "4.5": {
                "en-US": "Combat Action: When your active character is Charlotte, equip this card.\nAfter Charlotte equips this card, immediately use Framing: Freezing Point Composition once.\nWhen your Charlotte, who has this card equipped, is on the field, and after any of your characters uses a Normal Attack: If your opponent has a character on the field with Snappy Silhouette attached, heal your active character for 2 HP. (Once per Round)\n(You must have Charlotte in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为夏洛蒂时，装备此牌。\n夏洛蒂装备此牌后，立刻使用一次取景·冰点构图法。\n装备有此牌的夏洛蒂在场时，我方角色进行普通攻击后：如果对方场上有角色附属有瞬时剪影，则治疗我方出战角色2点。（每回合1次）\n（牌组中包含夏洛蒂，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Charlotte.png",  # noqa: E501
        "id": 211101,
    },
}


register_class(
    Charlotte_4_5 | SnappySilhoete_4_5 | NewsflashField_4_5 | ASummationOfInterest_4_5,
    character_descs,
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Character status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Skills


class FireworkFlareUp(PhysicalNormalAttackBase):
    name: Literal["Firework Flare-Up"] = "Firework Flare-Up"
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.PYRO)


class NiwabiFireDance(ElementalSkillBase):
    name: Literal["Niwabi Fire-Dance"] = "Niwabi Fire-Dance"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=1)

    def get_actions(self, match: Any) -> List[CreateObjectAction]:
        """
        only create object
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        talent = character.talent
        max_usage = 2
        if talent is not None:
            max_usage = talent.status_max_usage
        return [
            self.create_character_status(
                "Niwabi Enshou", {"usage": max_usage, "max_usage": max_usage}
            )
        ]


class RyuukinSaxifrage(ElementalBurstBase):
    name: Literal["Ryuukin Saxifrage"] = "Ryuukin Saxifrage"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [self.create_team_status("Aurous Blaze")]


# Talents


class NaganoharaMeteorSwarm_4_2(SkillTalent):
    name: Literal["Naganohara Meteor Swarm"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Yoimiya"] = "Yoimiya"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO,
        elemental_dice_number=2,
    )
    skill: Literal["Niwabi Fire-Dance"] = "Niwabi Fire-Dance"
    status_max_usage: int = 3


# character base


class Yoimiya_3_8(CharacterBase):
    name: Literal["Yoimiya"]  # Do not set default value for character name
    version: Literal["3.8"] = "3.8"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[FireworkFlareUp | NiwabiFireDance | RyuukinSaxifrage] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [FireworkFlareUp(), NiwabiFireDance(), RyuukinSaxifrage()]


register_class(Yoimiya_3_8 | NaganoharaMeteorSwarm_4_2)



# This file will record minimum supported version of frontend and current version.
# When frontend is updated, modify the version here.
from .. import __version_tuple__, __version__

__frontend_version__ = "0.4.5.0"


__all__ = [
    "__frontend_version__",
    "__version__",
    "__version_tuple__",
]



from typing import Literal

from ...utils.class_registry import register_class
from .base import AttackerSummonBase
from ..consts import DamageElementalType


class BurningFlame_3_3(AttackerSummonBase):
    name: Literal["Burning Flame"] = "Burning Flame"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1
    renew_type: Literal["ADD"] = "ADD"


register_class(BurningFlame_3_3)



from typing import Dict, List, Literal

from ....event import (
    ReceiveDamageEventArguments,
    RoundPrepareEventArguments,
    UseSkillEventArguments,
)
from .....utils.class_registry import register_class
from ....status.character_status.base import DefendCharacterStatus
from ....summon.base import AttackAndGenerateStatusSummonBase
from ....action import Actions, ChargeAction, CreateObjectAction
from ....match import Match
from ....struct import Cost
from ....consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    PassiveSkillBase,
    SkillBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class PyroScorpionGuardianStance_4_3(DefendCharacterStatus):
    name: Literal["Pyro Scorpion: Guardian Stance"] = "Pyro Scorpion: Guardian Stance"
    version: Literal["4.3"] = "4.3"
    desc: Literal["", "talent"] = ""

    # usage and max usage will be set by summon
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1


class SpiritOfOmenPyroScorpion_4_3(AttackAndGenerateStatusSummonBase):
    name: Literal["Spirit of Omen: Pyro Scorpion"] = "Spirit of Omen: Pyro Scorpion"
    desc: Literal["", "talent"] = ""
    version: Literal["4.3"] = "4.3"
    usage: int = 2
    max_usage: int = 2
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    status_name: str = "Pyro Scorpion: Guardian Stance"

    def renew(self, obj: "SpiritOfOmenPyroScorpion_4_3") -> None:
        super().renew(obj)
        if obj.desc == "talent" and self.desc != "talent":
            self.desc = "talent"

    def _create_status(self, match: Match) -> List[CreateObjectAction]:
        """
        Create for eremite self
        """
        characters = match.player_tables[self.position.player_idx].characters
        for c in characters:
            if c.name == "Eremite Scorching Loremaster":
                usage = 1
                if self.desc == "talent":
                    usage = 2
                return [
                    CreateObjectAction(
                        object_name="Pyro Scorpion: Guardian Stance",
                        object_position=c.position.set_area(
                            ObjectPositionType.CHARACTER_STATUS
                        ),
                        object_arguments={
                            "desc": self.desc,
                            "usage": usage,
                            "max_usage": usage,
                        },
                    )
                ]
        else:
            raise AssertionError("Eremite Scorching Loremaster not found")

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        Reset damage and call super
        """
        self.damage = 1
        return super().event_handler_ROUND_PREPARE(event, match)

    def event_handler_USE_SKILL(
        self, event: UseSkillEventArguments, match: Match
    ) -> List[Actions]:
        """
        If talent activated, and self eremite use normal or elemental skill,
        damage be 2.
        """
        if self.desc != "talent":
            # not talent, do nothing
            return []
        skill_position = event.action.skill_position
        if skill_position.player_idx != self.position.player_idx:
            # not self character use skill, do nothing
            return []
        character = match.player_tables[skill_position.player_idx].characters[
            skill_position.character_idx
        ]
        if character.name != "Eremite Scorching Loremaster":
            # not eremite use skill, do nothing
            return []
        skill: SkillBase = match.get_object(skill_position)  # type: ignore
        assert skill is not None
        if (
            skill.skill_type == SkillType.NORMAL_ATTACK
            or skill.skill_type == SkillType.ELEMENTAL_SKILL
        ):
            # is self eremite use normal or elemental skill, damage be 2
            self.damage = 2
        return []


class SpiritOfOmensAwakeningPyroScorpion(ElementalBurstBase):
    name: Literal[
        "Spirit of Omen's Awakening: Pyro Scorpion"
    ] = "Spirit of Omen's Awakening: Pyro Scorpion"
    damage: int = 2
    damage_type: Literal[DamageElementalType.PYRO] = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        summon_desc = ""
        if self.is_talent_equipped(match):
            summon_desc = "talent"
        return super().get_actions(match) + [
            self.create_summon("Spirit of Omen: Pyro Scorpion", {"desc": summon_desc})
        ]


class SpiritOfOmensPower(PassiveSkillBase):
    name: Literal["Spirit of Omen's Power"] = "Spirit of Omen's Power"
    triggered: bool = False

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[ChargeAction]:
        """
        When self receive damage, and self alive, and self hp <= 7,
        and not triggered, gain 1 charge
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if not event.final_damage.is_corresponding_character_receive_damage(
            character.position, match, ignore_piercing=False
        ):
            # not self receive damage, do nothing
            return []
        if character.is_defeated:  # pragma: no cover
            # self defeated, do nothing
            return []
        if character.hp > 7:
            # self hp > 7, do nothing
            return []
        if self.triggered:
            # already triggered, do nothing
            return []
        self.triggered = True
        return [
            ChargeAction(
                player_idx=character.position.player_idx,
                character_idx=character.position.character_idx,
                charge=1,
            )
        ]


class Scorpocalypse_4_3(SkillTalent):
    name: Literal["Scorpocalypse"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal[
        "Eremite Scorching Loremaster"
    ] = "Eremite Scorching Loremaster"
    skill: Literal[
        "Spirit of Omen's Awakening: Pyro Scorpion"
    ] = "Spirit of Omen's Awakening: Pyro Scorpion"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=2
    )


class EremiteScorchingLoremaster_4_3(CharacterBase):
    name: Literal["Eremite Scorching Loremaster"]
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase
        | ElementalSkillBase
        | SpiritOfOmensAwakeningPyroScorpion
        | SpiritOfOmensPower
    ] = []
    faction: List[FactionType] = [FactionType.THE_EREMITES]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Searing Glare",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name="Blazing Strike",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalSkillBase.get_cost(self.element),
            ),
            SpiritOfOmensAwakeningPyroScorpion(),
            SpiritOfOmensPower(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Eremite Scorching Loremaster": {
        "names": {
            "en-US": "Eremite Scorching Loremaster",
            "zh-CN": "镀金旅团·炽沙叙事人",
        },
        "image_path": "cardface/Char_Monster_Muscleman.png",  # noqa: E501
        "id": 2303,
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
    },
    "SKILL_Eremite Scorching Loremaster_NORMAL_ATTACK/Searing Glare": {
        "names": {"en-US": "Searing Glare", "zh-CN": "烧蚀之光"},
        "descs": {
            "4.3": {"en-US": "Deals 1 Pyro DMG.", "zh-CN": "造成1点火元素伤害。"}
        },
    },
    "SKILL_Eremite Scorching Loremaster_ELEMENTAL_SKILL/Blazing Strike": {
        "names": {"en-US": "Blazing Strike", "zh-CN": "炎晶迸击"},
        "descs": {
            "4.3": {"en-US": "Deals 3 Pyro DMG.", "zh-CN": "造成3点火元素伤害。"}
        },
    },
    "SKILL_Eremite Scorching Loremaster_ELEMENTAL_BURST/Spirit of Omen's Awakening: Pyro Scorpion": {  # noqa: E501
        "names": {
            "en-US": "Spirit of Omen's Awakening: Pyro Scorpion",
            "zh-CN": "厄灵苏醒·炎之魔蝎",
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Pyro DMG, summons 1 Spirit of Omen: Pyro Scorpion.",  # noqa: E501
                "zh-CN": "造成2点火元素伤害，召唤厄灵·炎之魔蝎。",
            }
        },
    },
    "SUMMON/Spirit of Omen: Pyro Scorpion": {
        "names": {"en-US": "Spirit of Omen: Pyro Scorpion", "zh-CN": "厄灵·炎之魔蝎"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deal 1 Pyro DMG.\nUsage(s): 2\n\nWhen entering play, when the Action Phase starts: Equips Pyro Scorpion: Guardian Stance to your Eremite Scorching Loremaster. (When Spirit of Omen: Pyro Scorpion is in play, the character will take 1 less DMG, once per Round.)",  # noqa: E501
                "zh-CN": "结束阶段：造成1点火元素伤害。\n可用次数：2\n\n入场时和行动阶段开始：使我方镀金旅团·炽沙叙事人附属炎之魔蝎·守势。（厄灵·炎之魔蝎在场时每回合1次，使角色受到的伤害-1。）",  # noqa: E501
            }
        },
        "image_path": "cardface/Summon_Muscleman.png",
    },
    "SUMMON/Spirit of Omen: Pyro Scorpion_talent": {
        "names": {"en-US": "Spirit of Omen: Pyro Scorpion", "zh-CN": "厄灵·炎之魔蝎"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deal 1 Pyro DMG; in Rounds when Eremite Scorching Loremaster has used a Normal Attack or Elemental Skill, the damage dealt +1.\nUsage(s): 2\n\nWhen entering play, when the Action Phase starts: Equips Pyro Scorpion: Guardian Stance to your Eremite Scorching Loremaster. (When Spirit of Omen: Pyro Scorpion is in play, the character will take 1 less DMG, twice per Round.)",  # noqa: E501
                "zh-CN": "结束阶段：造成1点火元素伤害；在镀金旅团·炽沙叙事人使用过「普通攻击」或「元素战技」的回合中，造成的伤害+1。\n可用次数：2\n\n入场时和行动阶段开始：使我方镀金旅团·炽沙叙事人附属炎之魔蝎·守势。（厄灵·炎之魔蝎在场时每回合2次，使角色受到的伤害-1。）",  # noqa: E501
            }
        },
        "image_path": "cardface/Summon_Muscleman.png",
    },
    "CHARACTER_STATUS/Pyro Scorpion: Guardian Stance": {
        "names": {"en-US": "Pyro Scorpion: Guardian Stance", "zh-CN": "炎之魔蝎·守势"},
        "descs": {
            "4.3": {
                "en-US": "the character will take 1 less DMG, once per Round.",  # noqa: E501
                "zh-CN": "使角色受到的伤害-1，每回合1次。",
            },
        },
    },
    "CHARACTER_STATUS/Pyro Scorpion: Guardian Stance_talent": {
        "names": {"en-US": "Pyro Scorpion: Guardian Stance", "zh-CN": "炎之魔蝎·守势"},
        "descs": {
            "4.3": {
                "en-US": "the character will take 1 less DMG, twice per Round.",  # noqa: E501
                "zh-CN": "使角色受到的伤害-1，每回合2次。",
            },
        },
    },
    "SKILL_Eremite Scorching Loremaster_PASSIVE/Spirit of Omen's Power": {
        "names": {"en-US": "Spirit of Omen's Power", "zh-CN": "厄灵之能"},
        "descs": {
            "4.3": {
                "en-US": "(Passive) After this character takes DMG: If this character has no greater than 7 HP, they gain 1 Energy. (Once per Match)",  # noqa: E501
                "zh-CN": "【被动】此角色受到伤害后：如果此角色生命值不多于7，则获得1点充能。（整场牌局限制1次）",  # noqa: E501
            }
        },
    },
    "TALENT_Eremite Scorching Loremaster/Scorpocalypse": {
        "names": {"en-US": "Scorpocalypse", "zh-CN": "魔蝎烈祸"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Eremite Scorching Loremaster, equip this card.\nAfter Eremite Scorching Loremaster equips this card, immediately use Spirit of Omen's Awakening: Pyro Scorpion once.\nSpirit of Omen: Pyro Scorpion created by Eremite Scorching Loremaster with this card equipped will deal +1 DMG in Rounds when Eremite Scorching Loremaster has used a Normal Attack or Elemental Skill.\nSpirit of Omen: Pyro Scorpion's DMG reduction effect can now be triggered twice per Round.\n(Your deck must contain Eremite Scorching Loremaster to add this card to it)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为镀金旅团·炽沙叙事人时，装备此牌。\n镀金旅团·炽沙叙事人装备此牌后，立刻使用一次厄灵苏醒·炎之魔蝎。\n装备有此牌的镀金旅团·炽沙叙事人生成的厄灵·炎之魔蝎在镀金旅团·炽沙叙事人使用过「普通攻击」或「元素战技」的回合中，造成的伤害+1；\n厄灵·炎之魔蝎的减伤效果改为每回合至多2次。\n（牌组中包含镀金旅团·炽沙叙事人，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Muscleman.png",  # noqa: E501
        "id": 223031,
    },
}


register_class(
    PyroScorpionGuardianStance_4_3
    | SpiritOfOmenPyroScorpion_4_3
    | Scorpocalypse_4_3
    | EremiteScorchingLoremaster_4_3,
    desc,
)



from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_pidx_cidx,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_dead_agent():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "TEST 2 p0 2 3 p1 2 3",
            "skill 0 0 1 2",
            "TEST 2 p0 0 3 p1 2 1",
            "TEST 3 p0c0 pyro",
            "card 0 0 0 1 2",
            "skill 1 0 1 2",
            "TEST 1 7 10 10 10 9 8 10 10",
            "skill 2 0 1 2",
            "sw_char 1 0",
            "TEST 1 7 9 10 10 4 8 10 10",
            "TEST 3 p1c1 pyro",
            "skill 1 0 1 2",
            "end",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "TEST 2 p0 1 3 p1 2 2",
            "TEST 1 10 10 10 10 10 8 10 10",
            "skill 0 0 1 2",
            "TEST 1 8 10 10 10 10 8 10 10",
            "TEST 2 p0 3 3 p1 2 0",
            "sw_char 0 0",
            "TEST 1 8 10 10 10 9 8 10 10",
            "skill 1 0 1 2",
            "TEST 1 7 10 10 10 4 8 10 10",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "sw_char 3 0",
            "end",
            "skill 1 0 1 2",
            "TEST 1 7 6 10 10 4 6 10 10",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fatui Pyro Agent*2
        character:Nahida
        character:Fischl
        Paid in Full*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    # equip talent initially for agent c1
    for table in match.player_tables:
        character = table.characters[1]
        assert character.name == "Fatui Pyro Agent"
        from lpsim.server.character.pyro.fatui_pyro_agent_3_3 import PaidinFull_3_3
        from lpsim.server.consts import ObjectPositionType
        from lpsim.server.struct import ObjectPosition

        talent = PaidinFull_3_3(name="Paid in Full")
        talent.position = ObjectPosition(
            player_idx=character.position.player_idx,
            character_idx=character.position.character_idx,
            area=ObjectPositionType.CHARACTER,
            id=talent.id,
        )
        character.attaches.append(talent)
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:4], hps[4:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.strip().split(" ")[3:]
                cmd = cmd[:2] + cmd[3:]
                data = [int(x) for x in cmd]
                data = [data[:2], data[2:]]
                for table, d in zip(match.player_tables, data):
                    for c, u in zip(table.characters, d):
                        if u == 0:
                            assert len(c.status) == 0
                            continue
                        assert c.status[0].usage == u
            elif test_id == 3:
                cmd = cmd.strip().split(" ")[2]
                pid = int(cmd[1])
                cid = int(cmd[3])
                assert match.player_tables[pid].characters[cid].element_application == [
                    "PYRO"
                ]
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_dead_agent_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "card 3 0",
            "card 2 0 12 11 10",
            "TEST 2 p0c0 usage 1 2",
            "TEST 2 p1c0 usage 2",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "TEST 2 p0c0 usage 1 3",
            "TEST 1 9 10 10 10 7 10 10 10",
            "skill 1 12 11 10",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fatui Pyro Agent*2
        character:Nahida
        character:Fischl
        Paid in Full*15
        Sweet Madame*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            cmd = cmd.strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:4], hps[4:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx, cidx = get_pidx_cidx(cmd)
                status = match.player_tables[pidx].characters[cidx].status
                check_usage(status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_dead_agent()
    test_dead_agent_2()



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import DamageIncreaseValue

from ....event import SkillEndEventArguments

from ....consts import ObjectPositionType, SkillType

from ....action import ChargeAction

from ....struct import Cost
from .base import ArtifactBase, RoundEffectArtifactBase


class OrnateKabuto_4_0(ArtifactBase):
    name: Literal["Ornate Kabuto"]
    version: Literal["4.0"] = "4.0"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 0

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChargeAction]:
        if not (
            self.position.area == ObjectPositionType.CHARACTER
            and event.action.position.player_idx == self.position.player_idx
            and event.action.position.character_idx != self.position.character_idx
            and event.action.skill_type == SkillType.ELEMENTAL_BURST
        ):
            # not equipped or not our other use burst
            return []
        # charge self by 1
        return [
            ChargeAction(
                player_idx=self.position.player_idx,
                character_idx=self.position.character_idx,
                charge=1,
            )
        ]


class OrnateKabuto_3_5(OrnateKabuto_4_0):
    version: Literal["3.5"]
    cost: Cost = Cost(any_dice_number=2)


class EmblemOfSeveredFate_4_1(OrnateKabuto_4_0, RoundEffectArtifactBase):
    name: Literal["Emblem of Severed Fate"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_per_round: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_BURST
        ):
            # not self use elemental burst
            return value
        if self.usage <= 0:
            # no usage left
            return value
        # modify damage
        assert mode == "REAL"
        self.usage -= 1
        value.damage += 2
        return value


class EmblemOfSeveredFate_4_0(EmblemOfSeveredFate_4_1):
    version: Literal["4.0"]
    max_usage_per_round: int = 999


class EmblemOfSeveredFate_3_7(EmblemOfSeveredFate_4_0):
    version: Literal["3.7"]
    cost: Cost = Cost(any_dice_number=3)


register_class(
    OrnateKabuto_4_0
    | EmblemOfSeveredFate_4_1
    | EmblemOfSeveredFate_4_0
    | EmblemOfSeveredFate_3_7
    | OrnateKabuto_3_5
)



import sys
from lpsim.tools import interact_match_with_agents, read_log
from lpsim.network import HTTPServer
from lpsim.server.match import MatchConfig
import logging


logging.basicConfig(level=logging.INFO)


if __name__ == "__main__":
    log_path = "logs.json"
    if len(sys.argv) > 1:
        log_path = sys.argv[1]
    log_str = open(log_path).read()
    agents, match = read_log(log_str, use_16_omni=None)
    # for i in range(2):
    #     characters = match.player_tables[i].player_deck_information.characters  # noqa: E501
    #     for c in characters:
    #         c.hp = c.max_hp = 30
    match.config.history_level = 10
    try:
        interact_match_with_agents(agents[0], agents[1], match)
    except Exception:
        print("*************************************************************")
        print("!!!!! ERROR: play log FAILED, play to last success log. !!!!!")
        print("*************************************************************")
    server = HTTPServer(
        decks=["", ""],
        match_config=MatchConfig(
            check_deck_restriction=False,
            card_number=None,
            max_same_card_number=None,
            character_number=None,
            max_round_number=999,
            random_first_player=False,
            history_level=10,
        ),
    )
    server.match = match

    server.run()



"""
Base class of all agents.
"""
from ..server.interaction import Responses
from ..server.match import Match
from ..utils import BaseModel


class AgentBase(BaseModel):
    """
    Base class of all agents.
    """

    player_idx: int

    def generate_response(self, match: Match) -> Responses | None:
        """
        generate response based on the match. Return None if no response is
        generated.
        """
        raise NotImplementedError()
